/*
 * Copyright (C) 2023 SUSE Software Solutions Germany GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 */

#include "config.h"

#include "grd-settings-handover.h"

struct _GrdSettingsHandover
{
  GrdSettings parent;
};

G_DEFINE_TYPE (GrdSettingsHandover, grd_settings_handover, GRD_TYPE_SETTINGS)

GrdSettingsHandover *
grd_settings_handover_new (void)
{
  return g_object_new (GRD_TYPE_SETTINGS_HANDOVER,
                       "runtime-mode", GRD_RUNTIME_MODE_HANDOVER,
                       "rdp-enabled", TRUE,
                       "rdp-view-only", FALSE,
                       "rdp-screen-share-mode", GRD_RDP_SCREEN_SHARE_MODE_EXTEND,
                       NULL);
}

static void
grd_settings_handover_init (GrdSettingsHandover *settings_handover)
{
}

static void
grd_settings_handover_class_init (GrdSettingsHandoverClass *klass)
{
}
